/**
 *    '$RCSfile: DelimiterResolver.java,v $'
 *
 *     '$Author: tao $'
 *       '$Date: 2004/08/12 21:10:14 $'
 *   '$Revision: 1.1 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.dataquery;

import java.util.Hashtable;

import org.apache.xpath.XPathAPI;

import org.ecoinformatics.util.Config;
import org.ecoinformatics.util.DebugPrinter;

import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

/**
 * This class will map delimter format in metadata (e.g eml documents) to the
 * format in db (e.g, hsql).
 * @author Jing Tao
 *
 */

public class DelimiterResolver
{
  private Hashtable specialMappingHash = new Hashtable();
  private DebugPrinter dbg             = DebugPrinter.getInstance();

  private static final String METADATAHEXAPREFIX  = "0x";
  private static final String DBHEXAPREFIX        =  "\\u";

  /**
   * Constructor. It will read config file and find the sepciy delimiter mapping
   * then put them into a hash table. metadata delimiter is a key and db delimter
   * is a value. For example, ; maps \semi.
   */
  public DelimiterResolver()
  {
    try
    {
      NodeList nl = new Config().getDelimiterMappings();
      for(int i=0; i<nl.getLength(); i++)
      {
        Node n = nl.item(i);
        Node metadataDelimiterNode =
                           XPathAPI.selectSingleNode(n, "./metadataDelimiter");
        Node dbDelimiterNode =
                        XPathAPI.selectSingleNode(n, "./DBdelimiter");
        if( metadataDelimiterNode == null ||  dbDelimiterNode == null )
        {
          throw new RuntimeException("Expecting a metadataDelimiter or " +
                                     "DBDelimiter node " +
                                      "child of mapping in " +
                                      "config.xml.");
        }

        String metadataDelimiter =
                          metadataDelimiterNode.getFirstChild().getNodeValue();
        String dbDelimiter =
                     dbDelimiterNode.getFirstChild().getNodeValue();
        dbg.print("The metadata delimiter " + metadataDelimiter +
                  " and db demliter" + dbDelimiter
                  + " will be put into mapping hash", 2 );
        specialMappingHash.put(metadataDelimiter, dbDelimiter);
      }//for

    }
    catch(javax.xml.transform.TransformerException te)
    {
      throw new RuntimeException("Error executing xpath expression in " +
        "DelimiterResolver.");
    }
    catch(Exception e)
    {
      throw new RuntimeException("Exception in DelimiterResolver.");
    }

  }// DelimiterResolver

  /**
   * This method will figure out a db delimiter when given a metadata delimiter.
   * If metadataDelimiter is null, null will be returned.
   * Here is way:
   * 1. If the given metadataDelimiter is in the specialhash (read from
   *    config file), the mapping will return from the hash.
   * 2. If the given metadataDelimter is a hexademical(start with "0x", e.g. 0x##),
   *    the "\\u##" will be returned.
   * 3. The others will return itself(the given metadataDelimiter)
   * @param metadataDelimiter String
   * @return String
   */
  public String resolve(String metadataDelimiter)
  {
    String dbDelimiter = null;
    if (metadataDelimiter == null)
    {
      return dbDelimiter;
    }

    if (specialMappingHash.containsKey(metadataDelimiter))
    {
      dbDelimiter = (String)specialMappingHash.get(metadataDelimiter);
    }
    else if (metadataDelimiter.startsWith(METADATAHEXAPREFIX))
    {
      dbDelimiter = DBHEXAPREFIX +
                       metadataDelimiter.substring(METADATAHEXAPREFIX.length());
    }
    else
    {
        dbDelimiter = metadataDelimiter;
    }
    dbg.print("The dbDelimiter is " + dbDelimiter, 2);
    return dbDelimiter;
  }

}
