/**
 *    '$RCSfile: TableEntity.java,v $'
 *
 *     '$Author: jones $'
 *       '$Date: 2004/05/28 23:04:27 $'
 *   '$Revision: 1.3 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.datasource;

import java.util.Vector;

/**
 * This object represents an TableEntity.  A TableEntity stores
 * information about a table of Attributes that is used in a Step
 * in the pipeline.
 */
public class TableEntity extends DataItemDescription
{
    /**static variable for ROWMAJOR tables**/
    public static String ROWMAJOR = "ROWMAJOR";
    /**static variable for COLUMNMAJOR tables**/
    public static String COLUMNMAJOR = "COLUMNMAJOR";

    /**static variable for table type**/
    public static String TABLEENTITY = "TABLEENTITY";

    private Vector attributes;
    private Boolean caseSensitive;
    private String orientation;
    private int numRecords = 0;
    private int numHeaderLines = -1;
    private String delimiter = null;
    private String recordDelimiter = null;
    private boolean multiple = false; // if true, multiple inputs can be mapped to one table

    private String fileName;    // filename where TableEntity data is stored

    /**
     * construct this object with some extra parameters
     * @param name the name of the tableEntity
     * @param description the description of the tableEntity
     * @param caseSensitive indicates whether this tableEntity is caseSensitive
     * or not.
     * @param orientation indicates whether this tableEntity is column or row
     * major
     * @param numRecords the number of records in this tableEntity
     */
    public TableEntity(String id, String name, String description,
                       Boolean caseSensitive, String orientation,
                       int numRecords)
    {
        this(id, name, description, null);
        if (caseSensitive != null) {
            this.caseSensitive = caseSensitive;
        }
        if (orientation != null) {
            this.orientation = orientation;
        }
        this.numRecords = numRecords;
    }

    /**
     * Construct a TableEntity, setting the list of attributes.
     */
    public TableEntity(String id, String name, String description,
            Attribute[] attributeList)
    {
        super(id, name, TABLEENTITY, description);
        fileName = "";
        attributes = new Vector();
        if (attributeList != null) {
            for (int i=0; i<attributeList.length; i++) {
                this.add(attributeList[i]);
            }
        }
        this.caseSensitive = new Boolean(false);
        this.orientation = "";
    }

    /**
     * Add an Attribute to this table.
     */
    public void add(Attribute a)
    {
      this.attributes.addElement(a);
      a.setParent(this);
    }

    /**
     * Return the unit for this TableEntity
     */
    public Attribute[] getAttributes()
    {
        Attribute[] atts = new Attribute[attributes.size()];
        return (Attribute[])this.attributes.toArray(atts);
    }

    /**
     * indicates whether the tableEntity is caseSensitive or not
     */
    public Boolean getCaseSensitive()
    {
      return caseSensitive;
    }

    /**
     * gets the orientation of the table entity
     */
    public String getOrientation()
    {
      return orientation;
    }

    /**
     * gets the number of records in the table entity
     */
    public int getNumRecords()
    {
      return numRecords;
    }

    /**
     * sets the number of header lines in the entity
     */
    public void setNumHeaderLines(int numHeaderLines)
    {
      this.numHeaderLines = numHeaderLines;
    }

    /**
     * get the number of header lines in the entity
     */
    public int getNumHeaderLines()
    {
      return this.numHeaderLines;
    }

    /**
     * set the delimiter used with this entity
     */
    public void setDelimiter(String delim)
    {
      this.delimiter = delim;
    }

    /**
     * get the delimiter used with this entity
     */
    public String getDelimiter()
    {
      return this.delimiter;
    }

    /**
     * set the record delimiter used with this entity
     */
    public void setRecordDelimiter(String delim)
    {
      this.recordDelimiter = delim;
    }

    /**
     * get the recordDelimiter used with this entity
     */
    public String getRecordDelimiter()
    {
      return this.recordDelimiter;
    }
    
    /*
    public String toString()
    {
      StringBuffer sb = new StringBuffer();
      sb.append("name: ").append(name).append("\n");
      sb.append("dataType: ").append(dataType).append("\n");
      sb.append("description: ").append(description).append("\n");
      sb.append("numRecords: ").append(numRecords).append("\n");
      sb.append("caseSensitive: ").append(caseSensitive).append("\n");
      sb.append("orientation: ").append(orientation).append("\n");
      sb.append("numHeaderLines: ").append(numHeaderLines).append("\n");
      sb.append("delimiter: ").append(delimiter).append("\n");
      sb.append("attributes: {");
      for(int i=0; i<attributes.size(); i++)
      {
        sb.append(((Attribute)attributes.elementAt(i)).toString());
      }
      sb.append("\n}");
      return sb.toString();
    }
    */

    /**
     * Returns the fileName.
     * @return String
     */
    public String getFileName()
    {
      return fileName;
    }

    /**
     * Sets the fileName.
     * @param fileName The fileName to set
     */
    public void setFileName(String fileName)
    {
      this.fileName = fileName;
    }

    /**
     * Serialize the data item in XML format.
     */
    public String toXml()
    {
        StringBuffer x = new StringBuffer();
        x.append("<table-entity id=\"");
        x.append(getId());
        x.append("\"");
        if( multiple == true )
        {
          x.append(" multiple=\"true\"");
        }
        x.append(">\n");
        appendElement(x, "entityName", getName());
        appendElement(x, "entityType", getDataType());
        appendElement(x, "entityDescription", getDefinition());
        Attribute[] atts = getAttributes();
        for (int i = 0; i < atts.length; i++) {
            x.append(atts[i].toXml());
        }
        x.append("</table-entity>\n");

        return x.toString();
    }

    /**
     * Sets the multiple to true.
     */
    public void setMultiple()
    {
      this.multiple = true;
    }

    /**
     * Returns multiple.
     * @return boolean
     */
    public boolean isMultiple()
    {
      return multiple;
    }

}
