/* An actor that reads binary files.



@Copyright (c) 2002-2003 The Regents of the University of California.

All rights reserved.



Permission is hereby granted, without written agreement and without

license or royalty fees, to use, copy, modify, and distribute this

software and its documentation for any purpose, provided that the

above copyright notice and the following two paragraphs appear in all

copies of this software.



IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY

FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES

ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF

THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF

SUCH DAMAGE.



THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,

INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF

MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE

PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF

CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,

ENHANCEMENTS, OR MODIFICATIONS.



                                                PT_COPYRIGHT_VERSION 2

                                                COPYRIGHTENDKEY

*/



package org.geon;



import ptolemy.actor.TypedIOPort;

import ptolemy.actor.lib.Source;

import ptolemy.data.ArrayToken;

import ptolemy.data.BooleanToken;

import ptolemy.data.IntToken;

import ptolemy.data.StringToken;

import ptolemy.data.UnsignedByteToken;

import ptolemy.data.Token;

import ptolemy.data.expr.FileParameter;

import ptolemy.data.expr.Parameter;

import ptolemy.data.type.ArrayType;

import ptolemy.data.type.BaseType;

import ptolemy.kernel.CompositeEntity;

import ptolemy.kernel.util.*;



import java.io.FileInputStream;

import java.io.IOException;

import java.net.URL;



//////////////////////////////////////////////////////////////////////////

//// BinaryFileReader

/**

This actor reads a file or URL and outputs a bytes array.



@author  Efrat Jaeger

@version $Id: BinaryFileReader.java,v 1.3 2004/07/01 23:32:13 higgins Exp $

@since Ptolemy II 3.0.2

*/

public class BinaryFileReader extends Source {



    /** Construct an actor with the given container and name.

     *  @param container The container.

     *  @param name The name of this actor.

     *  @exception IllegalActionException If the actor cannot be contained

     *   by the proposed container.

     *  @exception NameDuplicationException If the container already has an

     *   actor with this name.

     */

    public BinaryFileReader(CompositeEntity container, String name)

            throws IllegalActionException, NameDuplicationException {

        super(container, name);



        output.setTypeEquals(new ArrayType(BaseType.UNSIGNED_BYTE));



        endOfFile = new TypedIOPort(this, "endOfFile", false, true);

        endOfFile.setTypeEquals(BaseType.BOOLEAN);



        fileOrURL = new FileParameter(this, "fileOrURL");



        _attachText("_iconDescription", "<svg>\n"

                + "<rect x=\"-25\" y=\"-20\" "

                + "width=\"50\" height=\"40\" "

                + "style=\"fill:white\"/>\n"

                + "<polygon points=\"-15,-10 -12,-10 -8,-14 -1,-14 3,-10"

                + " 15,-10 15,10, -15,10\" "

                + "style=\"fill:red\"/>\n"

                + "</svg>\n");

    }



    ///////////////////////////////////////////////////////////////////

    ////                     ports and parameters                  ////



    /** An output port that produces <i>false</i> until the end of file

     *  is reached, at which point it produces <i>true</i>. The type

     *  is boolean.

     */

    public TypedIOPort endOfFile;



    /** The file name or URL from which to read.  This is a string with

     *  any form accepted by FileParameter.

     *  @see FileParameter

     */

    public FileParameter fileOrURL;



    ///////////////////////////////////////////////////////////////////

    ////                         public methods                    ////



    /** In the case of <i>fileOrURL</i>, do nothing if the file name is

     * the same as the previous value of this attribute.

     *  @param attribute The attribute that has changed.

     *  @exception IllegalActionException If the specified attribute

     *   is <i>fileOrURL</i> and the file cannot be opened, or the previously

     *   opened file cannot be closed; or if the attribute is

     */

    public void attributeChanged(Attribute attribute)

            throws IllegalActionException {

        if (attribute == fileOrURL) {

            // NOTE: We do not want to close the file if the file

            // has not in fact changed.  We check this by just comparing

            // name, which is not perfect...

            if (_previousFileOrURL != null

                    && !fileOrURL.getExpression().equals(_previousFileOrURL)) {

                _previousFileOrURL = fileOrURL.getExpression();

                fileOrURL.close();

                // Ignore if the fileOrUL is blank.

                if (fileOrURL.getExpression().trim().equals("")) {

                    _reader = null;

                } else {

                    try {

                        _reader = new FileInputStream(fileOrURL.asFile());

                    } catch (IOException ioex) {

                          throw new IllegalActionException(this, ioex, "Error opening stream for reading");

                    }

                }

                _reachedEOF = false;

            }

        } else {

            super.attributeChanged(attribute);

        }

    }



    /** Clone the actor into the specified workspace.

     *  @return A new actor.

     *  @exception CloneNotSupportedException If a derived class contains

     *   an attribute that cannot be cloned.

     */

    public Object clone(Workspace workspace)

            throws CloneNotSupportedException {

        BinaryFileReader newObject = (BinaryFileReader)super.clone(workspace);

        newObject.nBytesRead = 0;

        newObject.bytesRead = null;

        newObject._reachedEOF = false;

        newObject._reader = null;

        return newObject;

    }



    /** Output the data read in the preinitialize() or in the previous

     *  invocation of postfire(), if there is any.

     *  @exception IllegalActionException If there's no director.

     */

    public void fire() throws IllegalActionException {

        super.fire();

        if (nBytesRead > 0) {

            Token _bytes[] = new Token[nBytesRead];

            for (int i = 0; i < nBytesRead; i++) {

                _bytes[i] = new UnsignedByteToken(bytesRead[i]);

            }

            output.send(0, new ArrayToken(_bytes));

        }

    }



    /** If this is called after prefire() has been called but before

     *  wrapup() has been called, then close any

     *  open file re-open it, and read the first chunk of bytes

     *  be produced in the next invocation of prefire(). This occurs if

     *  this actor is re-initialized during a run of the model.

     *  @exception IllegalActionException If the file or URL cannot be

     *   opened or read.

     */

    public void initialize() throws IllegalActionException {

        super.initialize();

        if (_firedSinceWrapup) {

            _reader = null;

            _openAndReadFirstBytes();

        }

    }



    /** Read the next bytes from the file. If there reached EOF,

     *  return false.  Otherwise, return whatever the superclass returns.

     *  @exception IllegalActionException If there is a problem reading

     *   the file.

     */

    public boolean postfire() throws IllegalActionException {

        if (_reader == null) {

            return false;

        }

        try {

            nBytesRead = _reader.read(bytesRead);

            if (nBytesRead == 0) {

                // In case the return value gets ignored by the domain:

                _reachedEOF = true;

                endOfFile.broadcast(BooleanToken.TRUE);

                return false;

            }

            endOfFile.broadcast(BooleanToken.FALSE);

            return super.postfire();

        } catch (IOException ex) {

            throw new IllegalActionException(this, ex, "Postfire failed");

        }

    }



    /** Return false if there is no more data available in the file.

     *  Otherwise, return whatever the superclass returns.

     *  @exception IllegalActionException If the superclass throws it.

     */

    public boolean prefire() throws IllegalActionException {

        _firedSinceWrapup = true;

        if (_reachedEOF) return false;

        else return super.prefire();

    }



    /** Open the file or URL, and read the first bytes to

     *  be sent out in the fire() method.

     *  This is done in preinitialize() so

     *  that derived classes can extract information from the file

     *  that affects information used in type resolution or scheduling.

     *  @exception IllegalActionException If the file or URL cannot be

     *   opened or read.

     */

    public void preinitialize() throws IllegalActionException {

        super.preinitialize();

        _openAndReadFirstBytes();

    }



    /** Close the reader if there is one.

     *  @exception IllegalActionException If an IO error occurs.

     */

    public void wrapup() throws IllegalActionException {

        _reader = null;

        _firedSinceWrapup = false;

    }



    ///////////////////////////////////////////////////////////////////

    ////                         protected members                 ////



    /** number of bytes read. */

    protected int nBytesRead;



    /** The current bytes read. */

    protected byte[] bytesRead = new byte[20000];



    /** The current reader for the input file. */

    protected FileInputStream _reader;



    ///////////////////////////////////////////////////////////////////

    ////                         private methods                   ////



    /** Open the file and read the first bytes.

     */

    private void _openAndReadFirstBytes() throws IllegalActionException {

        try {

            _reader = new FileInputStream(fileOrURL.asFile());

        } catch (IOException ioex) {

            throw new IllegalActionException(this, ioex, "Error opening stream for reading");

        }

        _reachedEOF = false;

        try {

            nBytesRead = _reader.read(bytesRead);

        } catch (IOException ex) {

            throw new IllegalActionException(this, ex,

                    "Preinitialize failed.");

        }

    }



    ///////////////////////////////////////////////////////////////////

    ////                         private members                   ////



    /** Indicator that the fire() method has been called, but wrapup

     *  has not.  That is, we are in the middle of a run.

     */

    private boolean _firedSinceWrapup = false;



    /** Previous value of fileOrURL parameter. */

    private String _previousFileOrURL;



    /** Indicator that we have reached the end of file. */

    private boolean _reachedEOF = false;

}

